// Desk Jockey.cpp

#include "stdafx.h"
#include "resource.h"

// Global handle to this module's instance.
HINSTANCE g_hInstance;

// Main window class name.
const LPTSTR g_classname = TEXT("DeskJockey");

// Main window control identifiers.
enum
{
	idLabel,
	idCurrent,
	idDesktops,
	idNew,
	idSwitch,
	idRun
};

// Forward declaration.
DWORD WINAPI MyThreadProc(LPVOID lpv);

// General error display function.
void DisplayError(HWND hwnd, LPTSTR context)
{
	TCHAR errmsg[1024];
	FormatMessage(
		FORMAT_MESSAGE_FROM_SYSTEM,
		NULL,
		GetLastError(),
		0,
		errmsg,
		1024,
		NULL);
	MessageBox(hwnd, errmsg, context, MB_ICONERROR | MB_OK);
}

// Read the controls in the New Desktop dialog box and build
// arguments to the CreateDesktop() API based on their values.
// Runs when DialogProc_NewDesktop() gets WM_COMMAND from OK button.
void GetCreateDesktopArgs(HWND hwnd,
						  LPTSTR name,
						  int chars,
						  DWORD & dwFlags,
						  DWORD & dwDesiredAccess)
{
	// Read the name that the user typed.
	GetWindowText(GetDlgItem(hwnd, IDC_NAME), name, chars);

	// Read the checkboxes and build the bitfields accordingly.
	if (SendMessage(GetDlgItem(hwnd, IDC_HOOK), BM_GETCHECK, 0, 0))
		dwFlags = DF_ALLOWOTHERACCOUNTHOOK;
	else
		dwFlags = 0;

	dwDesiredAccess = 0;
	if (SendMessage(GetDlgItem(hwnd, IDC_CREATEMENU),
			BM_GETCHECK, 0, 0))
		dwDesiredAccess |= DESKTOP_CREATEMENU;
	if (SendMessage(GetDlgItem(hwnd, IDC_CREATEWINDOW),
			BM_GETCHECK, 0, 0))
		dwDesiredAccess |= DESKTOP_CREATEWINDOW;
	if (SendMessage(GetDlgItem(hwnd, IDC_ENUMERATE),
			BM_GETCHECK, 0, 0))
		dwDesiredAccess |= DESKTOP_ENUMERATE;
	if (SendMessage(GetDlgItem(hwnd, IDC_HOOKCONTROL),
			BM_GETCHECK, 0, 0))
		dwDesiredAccess |= DESKTOP_HOOKCONTROL;
	if (SendMessage(GetDlgItem(hwnd, IDC_JOURNALPLAY),
			BM_GETCHECK, 0, 0))
		dwDesiredAccess |= DESKTOP_JOURNALPLAYBACK;
	if (SendMessage(GetDlgItem(hwnd, IDC_JOURNALREC),
			BM_GETCHECK, 0, 0))
		dwDesiredAccess |= DESKTOP_JOURNALRECORD;
	if (SendMessage(GetDlgItem(hwnd, IDC_READOBJ),
			BM_GETCHECK, 0, 0))
		dwDesiredAccess |= DESKTOP_READOBJECTS;
	if (SendMessage(GetDlgItem(hwnd, IDC_SWITCH),
			BM_GETCHECK, 0, 0))
		dwDesiredAccess |= DESKTOP_SWITCHDESKTOP;
	if (SendMessage(GetDlgItem(hwnd, IDC_WRITEOBJ),
			BM_GETCHECK, 0, 0))
		dwDesiredAccess |= DESKTOP_WRITEOBJECTS;
}

// Dialog procedure for the New Desktop dialog.
// Runs when MainWindowProc() gets WM_COMMAND from New button.
INT_PTR CALLBACK DialogProc_NewDesktop(HWND hwnd,
									   UINT msg,
									   WPARAM wparam,
									   LPARAM lparam)
{
	int id, notification;

	switch(msg)
	{
	case WM_INITDIALOG:
		// Make all the checkboxes checked by default.
		SendMessage(GetDlgItem(hwnd, IDC_HOOK),
			BM_SETCHECK, BST_CHECKED, 0);
		SendMessage(GetDlgItem(hwnd, IDC_CREATEMENU),
			BM_SETCHECK, BST_CHECKED, 0);
		SendMessage(GetDlgItem(hwnd, IDC_CREATEWINDOW),
			BM_SETCHECK, BST_CHECKED, 0);
		SendMessage(GetDlgItem(hwnd, IDC_ENUMERATE),
			BM_SETCHECK, BST_CHECKED, 0);
		SendMessage(GetDlgItem(hwnd, IDC_HOOKCONTROL),
			BM_SETCHECK, BST_CHECKED, 0);
		SendMessage(GetDlgItem(hwnd, IDC_JOURNALPLAY),
			BM_SETCHECK, BST_CHECKED, 0);
		SendMessage(GetDlgItem(hwnd, IDC_JOURNALREC),
			BM_SETCHECK, BST_CHECKED, 0);
		SendMessage(GetDlgItem(hwnd, IDC_READOBJ),
			BM_SETCHECK, BST_CHECKED, 0);
		SendMessage(GetDlgItem(hwnd, IDC_SWITCH),
			BM_SETCHECK, BST_CHECKED, 0);
		SendMessage(GetDlgItem(hwnd, IDC_WRITEOBJ),
			BM_SETCHECK, BST_CHECKED, 0);

		// Put the cursor in the Name edit-box.
		SetFocus(GetDlgItem(hwnd, IDC_NAME));
		return FALSE;

	case WM_COMMAND:
		id = LOWORD(wparam);
		notification = HIWORD(wparam);
		switch(id)
		{
		case IDOK:
			// Attempt to create the new desktop.
			{
				TCHAR name[MAX_PATH];
				DWORD dwFlags, dwDesiredAccess;

				// Construct the args to CreateDesktop() based on
				// the values of each control in the dialog box.
				GetCreateDesktopArgs(
					hwnd,
					name,
					MAX_PATH,
					dwFlags,
					dwDesiredAccess);

				HDESK hdesk = CreateDesktop(
					name,
					NULL,
					NULL,
					dwFlags,
					dwDesiredAccess,
					NULL);

				// If the desktop was created successfully, start
				// a new thread to make our UI appear on it.
				// Pass a handle to the new desktop to the thread.
				if (hdesk != NULL)
				{
					HANDLE hthread = CreateThread(
						NULL,
						0,
						MyThreadProc,
						hdesk,
						0,
						NULL);
					CloseHandle(hthread);
				}
				else
				{
					DisplayError(hwnd, TEXT("CreateDesktop"));
				}

				EndDialog(hwnd, reinterpret_cast<INT_PTR>(hdesk));
			}
			break;

		case IDCANCEL:
			EndDialog(hwnd, 0);
			break;
		}
		break;

	case WM_CLOSE:
		EndDialog(hwnd, 0);
		break;
	}

	return FALSE;
}

// Start a command prompt on any named desktop.
// Called by MainWindowProc() upon WM_COMMAND from Run button.
void CreateNewShell(LPTSTR desktop)
{
	TCHAR path[MAX_PATH + 1];
	GetSystemDirectory(path, MAX_PATH);
	PathAppend(path, TEXT("cmd.exe")); // useful func in shlwapi.dll

	PROCESS_INFORMATION pi = {0};
	STARTUPINFO si = {0};

	si.cb = sizeof(si);
	si.lpTitle = desktop;
	si.lpDesktop = desktop; // here's the important line!

	if (CreateProcess(
		path,
		TEXT("cmd"),
		NULL,
		NULL,
		TRUE,
		CREATE_NEW_CONSOLE,
		NULL,
		NULL,
		&si,
		&pi))
	{
		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);
	}
	else
	{
		DisplayError(NULL, TEXT("CreateProcess"));
	}
}

// See which desktop is currently selected in the listbox.
// This function may be called by multiple threads, so use TLS!
LPTSTR ReadListboxSelection(HWND hwndList)
{
	// If this is the first time we're called,
	// take care of initialization.
	static DWORD dwSlot = -1;
	if (dwSlot == -1)
		dwSlot = TlsAlloc();

	// See if we allocated a string previously.
	// If so, delete it.
	LPTSTR text = static_cast<LPTSTR>(TlsGetValue(dwSlot));
	if (text != NULL)
	{
		delete [] text;
		text = NULL;
	}

	// We may get a NULL hwndList if we're just cleaning up.
	if (IsWindow(hwndList))
	{
		// Get the current selection, if any.
		int sel = SendMessage(hwndList, LB_GETCURSEL, 0, 0);
		if (sel != LB_ERR)
		{
			int chars = SendMessage(hwndList, LB_GETTEXTLEN, sel, 0);
			if (chars > 0)
			{
				text = new TCHAR[chars + 1];
				SendMessage(hwndList, LB_GETTEXT, sel,
					reinterpret_cast<LPARAM>(text));
			}
		}
	}

	// Save the text we just allocated, so we can delete it later.
	TlsSetValue(dwSlot, text);

	return text;
}

// Take each desktop and add it to the listbox.
BOOL CALLBACK MyDesktopEnumProc(LPTSTR name, LPARAM listbox)
{
	HWND hwndList = (HWND)listbox;
	SendMessage(hwndList, LB_ADDSTRING, 0,
		reinterpret_cast<LPARAM>(name));
	return TRUE;
}

// Reconstruct the listbox with updated list of desktops.
void UpdateListbox(HWND hwndList)
{
	// Save the current selection.
	LPTSTR text = ReadListboxSelection(hwndList);

	// Erase listbox.
	SendMessage(hwndList, LB_RESETCONTENT, 0, 0);

	// Populate listbox with desktop names.
	HWINSTA station = GetProcessWindowStation();
	EnumDesktops(
		station,
		MyDesktopEnumProc,
		(LPARAM)hwndList
		);
	CloseWindowStation(station);

	// Restore selection, if possible.
	if (text != NULL)
	{
		SendMessage(hwndList, LB_SELECTSTRING, 0,
			reinterpret_cast<LPARAM>(text));
	}
}

// Redraw the Switch and Run buttons with the name of the desktop
// currently selected in the listbox.
void UpdateButtons(LPTSTR text, HWND hwndSwitch, HWND hwndRun)
{
	if (text != NULL)
	{
		TCHAR buf[1024];

		EnableWindow(hwndSwitch, TRUE);
		wsprintf(buf, TEXT("Switch to %s"), text);
		SetWindowText(hwndSwitch, buf);

		EnableWindow(hwndRun, TRUE);
		wsprintf(buf, TEXT("Run CMD on %s"), text);
		SetWindowText(hwndRun, buf);
	}
	else
	{
		EnableWindow(hwndSwitch, FALSE);
		SetWindowText(hwndSwitch, TEXT("Switch"));

		EnableWindow(hwndRun, FALSE);
		SetWindowText(hwndRun, TEXT("Run CMD"));
	}
}

// Add the controls to the main window.
// Called by MainWindowProc() upon WM_CREATE.
void CreateChildControls(HWND hwnd)
{
	CreateWindow(
		TEXT("STATIC"),
		TEXT("This window's desktop:"),
		WS_CHILD | WS_VISIBLE,
		10, 10, 250, 20,
		hwnd,
		(HMENU) idLabel,
		g_hInstance,
		NULL
		);

	// Read the name of the desktop that this thread is using.
	// Note that GetUserObjectInformation() deals in bytes,
	// not in characters.  For a Unicode app, this matters!
	HDESK hdesk = GetThreadDesktop(GetCurrentThreadId());
	DWORD bytes;
	GetUserObjectInformation(hdesk, UOI_NAME, NULL, 0, &bytes);
	LPTSTR name = static_cast<LPTSTR>
		(HeapAlloc(GetProcessHeap(), 0, bytes));
	GetUserObjectInformation(hdesk, UOI_NAME, name, bytes, &bytes);
	HWND hwndCurrent = CreateWindow(
		TEXT("STATIC"),
		name,
		WS_CHILD | WS_VISIBLE | SS_SUNKEN,
		10, 30, 250, 25,
		hwnd,
		(HMENU) idCurrent,
		g_hInstance,
		NULL
		);
	HeapFree(GetProcessHeap(), 0, name);

	CreateWindow(
		TEXT("STATIC"),
		TEXT("Available desktops:"),
		WS_CHILD | WS_VISIBLE,
		10, 70, 250, 20,
		hwnd,
		(HMENU) idLabel,
		g_hInstance,
		NULL
		);

	HWND hwndList = CreateWindow(
		TEXT("LISTBOX"),
		TEXT(""),
		WS_CHILD | WS_VISIBLE | WS_VSCROLL | LBS_STANDARD,
		10, 100, 250, 150,
		hwnd,
		(HMENU) idDesktops,
		g_hInstance,
		NULL
		);
	UpdateListbox(hwndList);

	CreateWindow(
		TEXT("BUTTON"),
		TEXT("New"),
		WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
		10, 260, 250, 30,
		hwnd,
		(HMENU) idNew,
		g_hInstance,
		NULL
		);

	HWND hwndSwitch = CreateWindow(
		TEXT("BUTTON"),
		TEXT("Switch"),
		WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON | WS_DISABLED,
		10, 300, 250, 30,
		hwnd,
		(HMENU) idSwitch,
		g_hInstance,
		NULL
		);

	HWND hwndRun = CreateWindow(
		TEXT("BUTTON"),
		TEXT("Run CMD"),
		WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON | WS_DISABLED,
		10, 340, 250, 30,
		hwnd,
		(HMENU) idRun,
		g_hInstance,
		NULL
		);

	UpdateButtons(NULL, hwndSwitch, hwndRun);
}

// Window procedure for main UI.  Note that this may be executed
// by multiple threads, so it must be thread-safe!
LRESULT CALLBACK MainWindowProc(HWND hwnd,
								UINT msg,
								WPARAM wparam,
								LPARAM lparam)
{
	const k_TimerID = 123;
	int id, notification;
	INT_PTR dialog_result;
	HWND hwndList = NULL;
	HWND hwndSwitch = NULL;
	HWND hwndRun = NULL;

	switch (msg)
	{
	case WM_CREATE:
		CreateChildControls(hwnd);  // Add controls to the form.
		SetTimer(hwnd, k_TimerID, 5000, NULL); // Update every 5 sec.
		break;

	case WM_COMMAND:
		id = LOWORD(wparam);
		notification = HIWORD(wparam);
		hwndList = GetDlgItem(hwnd, idDesktops);
		hwndSwitch = GetDlgItem(hwnd, idSwitch);
		hwndRun = GetDlgItem(hwnd, idRun);

		switch(id)
		{
		case idDesktops:
			// User selected a different desktop.  Update buttons.
			if (notification == LBN_SELCANCEL ||
				notification == LBN_SELCHANGE)
			{
				LPTSTR text = ReadListboxSelection(hwndList);
				UpdateButtons(text, hwndSwitch, hwndRun);
			}
			break;

		case idNew:
			// User clicked the New button.
			dialog_result = DialogBox(
				g_hInstance,
				MAKEINTRESOURCE(IDD_NEWDESKTOP),
				hwnd,
				(DLGPROC)DialogProc_NewDesktop);
			if (dialog_result != 0)
			{
				UpdateListbox(hwndList);
				UpdateButtons(ReadListboxSelection(hwndList),
					hwndSwitch, hwndRun);
			}
			break;

		case idSwitch:
			// User clicked the Switch button.
			{
				LPTSTR text = ReadListboxSelection(hwndList);
				if (text == NULL)
					break;

				HDESK hdesk = OpenDesktop(
					text,
					0,
					FALSE,
					DESKTOP_SWITCHDESKTOP);
				if (hdesk != NULL)
				{
					SwitchDesktop(hdesk);
					CloseDesktop(hdesk);
				}
				else
				{
					DisplayError(hwnd, TEXT("OpenDesktop"));
				}
			}
			break;

		case idRun:
			// User clicked the Run button.
			CreateNewShell(ReadListboxSelection(hwndList));
			break;
		}
		break;

	case WM_TIMER:
		// Desktops may have been created by other apps, or may
		// have gone away, so rebuild desktop list and buttons.
		hwndList = GetDlgItem(hwnd, idDesktops);
		hwndSwitch = GetDlgItem(hwnd, idSwitch);
		hwndRun = GetDlgItem(hwnd, idRun);

		UpdateListbox(hwndList);
		UpdateButtons(ReadListboxSelection(hwndList),
			hwndSwitch, hwndRun);
		break;

	case WM_DESTROY:
		KillTimer(hwnd, k_TimerID); // No more updates needed.
		ReadListboxSelection(NULL); // Prevent memory leak.
		PostQuitMessage(0); // Exit thread (or app, if 1st thread).
		break;

	default:
		return DefWindowProc(hwnd, msg, wparam, lparam);
	}
	return 0;
}

void RegisterMyWindowClass()
{
	WNDCLASSEX wcx = {0};

	wcx.cbSize = sizeof(wcx);
	wcx.lpszClassName = g_classname;
	wcx.style = CS_SAVEBITS;
	wcx.lpfnWndProc = MainWindowProc;
	wcx.hInstance = g_hInstance;
	wcx.hIcon = LoadIcon(
		g_hInstance,
		MAKEINTRESOURCE(IDI_DESKJOCKEY));
	wcx.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcx.hIconSm = LoadIcon(
		g_hInstance,
		MAKEINTRESOURCE(IDI_DESKJOCKEY));
	wcx.hbrBackground = (HBRUSH)(COLOR_3DFACE + 1);

	RegisterClassEx(&wcx);
}

void CreateMyMainWindow()
{
	CreateWindowEx(
		WS_EX_OVERLAPPEDWINDOW,
		g_classname,
		TEXT("Desk Jockey"),
		WS_VISIBLE | WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU |
			WS_THICKFRAME | WS_MINIMIZEBOX,
		CW_USEDEFAULT, CW_USEDEFAULT, 280, 420,
		NULL,
		NULL,
		g_hInstance,
		NULL
		);
}

// Entry point for threads other than the main thread.
DWORD WINAPI MyThreadProc(LPVOID lpv)
{
	// Save handle to original desktop.
	HDESK hdeskOriginal = GetThreadDesktop(GetCurrentThreadId());

	// Start working with private desktop.
	HDESK hdesk = static_cast<HDESK>(lpv);
	if (SetThreadDesktop(hdesk))
	{
		CreateMyMainWindow();

		MSG msg;
		while (GetMessage(&msg, NULL, 0, 0) > 0)
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	// Restore original desktop, which must still be alive
	// precisely because it is this process's original desktop.
	SetThreadDesktop(hdeskOriginal);

	// Release private desktop.  It will be destroyed now, unless
	// some other process is still using it.
	CloseDesktop(hdesk);
	return 0;
}

// Entry point for the main thread.
int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
	g_hInstance = hInstance;

	RegisterMyWindowClass();
	CreateMyMainWindow();

	MSG msg;
	while (GetMessage(&msg, NULL, 0, 0) > 0)
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	return 0;
}